#ifndef __probes_h__
#define __probes_h__

#include "packet.h"

class Proxy;

class Hashtable {
protected:
  const static int maxEntries = 3000;
  struct entryStruct {
    int key;
    void *value;
  } entry[maxEntries];
  
public:
  Hashtable();
  void put(int key, void *value);
  void *get(int key);
};

class ProbeTrace {
protected:
  const static int maxReceived = 5000;
  int probesSent;
  int responsesReceived;
  int *receiveSequence;
  int *ipidSequence;
    
public:
  ProbeTrace();
  ProbeTrace(const char *moutFileName);
  ~ProbeTrace();
  void probeSent(int id);
  void responseReceived(int id, int ipid);
  void snapshot(int *probesSent, int *responsesReceived, int **receiveSequence, int **ipidSequence);
  void dump(FILE *stream);
};

class ProbePolicy {
protected:
  ProbeTrace *probeTrace;
  Proxy *proxy;
  
public:
  ProbePolicy() { this->probeTrace = NULL; this->proxy = NULL; };
  virtual ~ProbePolicy() {};
  virtual void receivePacketFromSender(Packet *packet) {};
  virtual void receivePacketFromReceiver(Packet *packet) {};
  virtual void receivePacketFromRemote(Packet *packet) {};
  virtual void setProbeTrace(ProbeTrace *probeTrace) { this->probeTrace = probeTrace; };
  virtual const char *getName() { return "unknown"; };
  void setProxy(Proxy *proxy) { this->proxy = proxy; };
};

class TcpAckPolicy : public ProbePolicy {
public:
  enum variationMode { VARY_NOTHING, VARY_SOURCE_PORT, VARY_DEST_PORT };

protected:
  const static int firstAckSeqNo = 7777;
  int packetsReceivedFromSender;
  int packetsReceivedFromRemote;
  int packetsSentToRemote;
  int nextAckIndex;
  Hashtable *packetTable;

  Packet *makeAck(Packet *original, uint32 idx);
  
public:
  TcpAckPolicy();
  virtual void receivePacketFromSender(Packet *packet);
  virtual void receivePacketFromReceiver(Packet *packet);
  virtual void receivePacketFromRemote(Packet *packet);
  virtual const char *getName() { return "tcp-ack"; };
};

struct IcmpEchoRequest {
  uint8 type;
  uint8 code;
  uint16 checksum;
  uint16 identifier;
  uint16 sequenceno;
  uint8 data[];
};

class IcmpEchoRequestPolicy : public ProbePolicy {
protected:
  Hashtable *packetTable;
  int nextSequenceNo;

  Packet *makeIcmpEchoRequest(Packet *packet, int sequenceNo);
  
public:
  IcmpEchoRequestPolicy();
  virtual void receivePacketFromSender(Packet *packet);
  virtual void receivePacketFromReceiver(Packet *packet);
  virtual void receivePacketFromRemote(Packet *packet);
  virtual const char *getName() { return "icmp-echo"; };
};

struct IcmpTimestampRequest {
  uint8 type;
  uint8 code;
  uint16 checksum;
  uint16 identifier;
  uint16 sequenceno;
  uint32 originateTime;
  uint32 receiveTime;
  uint32 transmitTime;
};

class IcmpTimestampRequestPolicy : public ProbePolicy {
protected:
  Hashtable *packetTable;
  int nextSequenceNo;

  Packet *makeIcmpTimestampRequest(Packet *packet, int sequenceNo);
  
public:
  IcmpTimestampRequestPolicy();
  virtual void receivePacketFromSender(Packet *packet);
  virtual void receivePacketFromReceiver(Packet *packet);
  virtual void receivePacketFromRemote(Packet *packet);
  virtual const char *getName() { return "icmp-timestamp"; };
};

class UdpDataPolicy : public ProbePolicy {
protected:
  Hashtable *packetTable;
  int nextSequenceNo;

  Packet *makeUdpDataPacket(Packet *packet, int sequenceNo);
  
public:
  UdpDataPolicy();
  virtual void receivePacketFromSender(Packet *packet);
  virtual void receivePacketFromReceiver(Packet *packet);
  virtual void receivePacketFromRemote(Packet *packet);
  virtual const char *getName() { return "udp-data"; };
};

#endif /* defined(__probes_h__) */
