#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <linux/netfilter.h>

#include "monarch.h"
#include "proxies.h"
#include "traces.h"

extern "C" {
#ifdef HAVE_LIBIPQ_H
#include <libipq.h>
#else
#include <libipq/libipq.h>
#endif
}

#define dprintf(a...) do { } while (0)
//#define dprintf(a...) printf(a)

NetfilterProxy::NetfilterProxy() : Proxy()
{
  this->initialized = false;
  this->captureSocket = -1;
  this->sendSocket = -1;
  this->rulesEstablished = 0;
  this->ipqHandle = NULL;
  this->rawSocket = -1;
  this->networkDeviceName[0] = 0;
  this->senderPort = -1;
  this->receiverPort = -1;
  memset(&localAddress, 0, sizeof(localAddress));
  memset(&remoteAddress, 0, sizeof(remoteAddress));
  memset(&senderAddress, 0, sizeof(senderAddress));
  memset(&receiverAddress, 0, sizeof(receiverAddress));
}

NetfilterProxy::~NetfilterProxy()
{
  removeFilterRules();
  if (rawSocket >= 0) {
    close(rawSocket);
    rawSocket = -1;
  }
}

void NetfilterProxy::getAddresses(struct sockaddr_in *senderLocal, struct sockaddr_in *senderRemote, struct sockaddr_in *receiverLocal, struct sockaddr_in *receiverRemote)
{
  if (senderLocal)
    *senderLocal = senderAddress;
  if (receiverLocal) 
    *receiverLocal = receiverAddress;
  if (senderRemote)
    *senderRemote = remoteAddress;
  if (receiverRemote)
    *receiverRemote = remoteAddress;
}

void NetfilterProxy::removeFilterRules()
{
  char buf1[256], buf2[256], rule[256];
  int retval;
  
  if (rulesEstablished > 2) {
    snprintf(rule, sizeof(rule) - 1,   
      "/sbin/iptables -D OUTPUT -p tcp -s %s -d %s --sport %d --dport %d -j QUEUE\n",  
      renderIP(ntohl(localAddress.sin_addr.s_addr), buf1), renderIP(ntohl(remoteAddress.sin_addr.s_addr), buf2),
      receiverPort, ntohs(remoteAddress.sin_port)
    ); 
    retval = system(rule);
    if (retval)
      fprintf(stderr, "NetfilterProxy: Cannot remote Netfilter rule: '%s' (exit code %d)", rule, retval);
    rulesEstablished --;
  }
  
  if (rulesEstablished > 1) {
    snprintf(rule, sizeof(rule) - 1,    
      "/sbin/iptables -D INPUT -p tcp -s %s -d %s --sport %d --dport %d -j QUEUE\n",   
      renderIP(ntohl(remoteAddress.sin_addr.s_addr), buf1), renderIP(ntohl(localAddress.sin_addr.s_addr), buf2),
      ntohs(remoteAddress.sin_port), ntohs(localAddress.sin_port)
    ); 

    retval = system(rule);
    if (retval)
      fprintf(stderr, "NetfilterProxy: Cannot remote Netfilter rule: '%s' (exit code %d)", rule, retval);
    rulesEstablished --;
  }
  
  if (rulesEstablished > 0) {
    snprintf(rule, sizeof(rule) - 1,   
      "/sbin/iptables -D OUTPUT -p tcp -s %s -d %s --sport %d --dport %d -j QUEUE\n", 
      renderIP(ntohl(localAddress.sin_addr.s_addr), buf1), renderIP(ntohl(remoteAddress.sin_addr.s_addr), buf2),
      senderPort, ntohs(remoteAddress.sin_port)
    );

    retval = system(rule);
    if (retval)
      fprintf(stderr, "NetfilterProxy: Cannot remote Netfilter rule: '%s' (exit code %d)", rule, retval);
    rulesEstablished --;
  }
}

void NetfilterProxy::sendRaw(Packet *p) 
{
  struct sockaddr_in sockAddr;
  int nbytes = 0;
  sockAddr.sin_family = AF_INET;
  sockAddr.sin_addr.s_addr = htonl(p->getDestination());
  sockAddr.sin_port = htons(p->getDestinationPort());
  struct timeval tv;
  gettimeofday(&tv, NULL);
  if (packetTrace)
    packetTrace->addPacket(&tv, p);
  nbytes = sendto(rawSocket, p->getBuffer(), p->getIpLength(), 0, (struct sockaddr *)&sockAddr, sizeof(sockAddr));
  if (nbytes < p->getIpLength())
    panic(ERR_CANTSEND, "NetfilterProxy: Tried to write %d bytes raw packet, but sendto returned %d", p->getIpLength(), nbytes);
}

void NetfilterProxy::forwardPacketToReceiver(Packet *packet)
{
  if (packet->isTcpPacket()) {
    uint8 *mssOption = packet->findTcpOption(Packet::TCP_OPTION_MSS);
    if (mssOption && (forcedMTU >= 0)) 
      *(uint16*)&mssOption[2] = htons(forcedMTU-(sizeof(struct IpHeader)+sizeof(struct TcpHeader)));
  }
  
  packet->setSourceAndDestination(&remoteAddress, &receiverAddress);
  sendRaw(packet);
}

void NetfilterProxy::forwardPacketToSender(Packet *packet)
{
  if (packet->isTcpPacket()) {
    uint8 *mssOption = packet->findTcpOption(Packet::TCP_OPTION_MSS);
    if (mssOption && (forcedMTU >= 0)) 
      *(uint16*)&mssOption[2] = htons(forcedMTU-(sizeof(struct IpHeader)+sizeof(struct TcpHeader)));
  }
  
  packet->setSourceAndDestination(&remoteAddress, &senderAddress);
  sendRaw(packet);
}

void NetfilterProxy::sendPacketToRemote(Packet *packet)
{
  packet->setSourceAndDestination(&localAddress, &remoteAddress);
  sendRaw(packet);
}

void NetfilterProxy::init(struct addressBlock *addressArg, const char *deviceName)
{
  char buf1[256], buf2[256], rule[256];
  int retval;

  assert(!initialized && !rulesEstablished && policy);
  
  if (geteuid() != 0)
    panic(ERR_PRIVILEGE, "NetfilterProxy: Must be run as user 'root'");
  if (strcmp(policy->getName(), "tcp-ack"))
    panic(ERR_PARAM, "NetfilterProxy: Probe types other than 'tcp-ack' not yet supported");
  if (addressArg->senderAddress.sin_family || addressArg->senderPeerAddress.sin_family || addressArg->receiverAddress.sin_family || addressArg->receiverPeerAddress.sin_family)
    panic(ERR_PARAM, "NetfilterProxy: Setting sender/receiver addresses not yet supported");
  
  localAddress = addressArg->localAddress;
  remoteAddress = addressArg->remoteAddress;
  strncpy(networkDeviceName, deviceName, sizeof(networkDeviceName));

  uint16 ports[2];
  findUnusedPorts(localAddress.sin_addr.s_addr, ports, 2);
  this->senderPort = ports[0];
  this->receiverPort = ports[1];

  senderAddress = localAddress;
  senderAddress.sin_port = htons(senderPort);
  receiverAddress = localAddress;
  receiverAddress.sin_port = htons(receiverPort);
  policy->setProxy(this);

  dprintf("Netfilter proxy starting with %s->%s\n", renderAddress(&localAddress, buf1), renderAddress(&remoteAddress, buf2));
  
  /* Open raw socket */

  if ((rawSocket = socket(AF_INET, SOCK_RAW, IPPROTO_RAW)) < 0)
    panic(ERR_INTERNAL, "NetfilterProxy: Cannot open raw socket (%s)", strerror(errno));
  
  int yes = 1;
  if (setsockopt(rawSocket, IPPROTO_IP, IP_HDRINCL, (char*)&yes, sizeof(yes)) < 0) 
    panic(ERR_INTERNAL, "NetfilterProxy: Cannot set IP_HDRINCL option on raw socket (%s)", strerror(errno));

  /* Insert capture rule: Sender -> Remote */
  
  snprintf(rule, sizeof(rule) - 1,   
    "/sbin/iptables -I OUTPUT -p tcp -s %s -d %s --sport %d --dport %d -j QUEUE\n", 
    renderIP(ntohl(localAddress.sin_addr.s_addr), buf1), renderIP(ntohl(remoteAddress.sin_addr.s_addr), buf2),
    senderPort, htons(remoteAddress.sin_port)
  );

  retval = system(rule);
  if (retval)
    panic(ERR_EXTCMD, "NetfilterProxy: Cannot create Netfilter rule: '%s' (exit code %d)", rule, retval);
  rulesEstablished ++;
    
  /* Insert capture rule: Remote -> Local */

  snprintf(rule, sizeof(rule) - 1,    
    "/sbin/iptables -I INPUT -p tcp -s %s -d %s --sport %d --dport %d -j QUEUE\n",   
    renderIP(ntohl(remoteAddress.sin_addr.s_addr), buf1), renderIP(ntohl(localAddress.sin_addr.s_addr), buf2),
    htons(remoteAddress.sin_port), htons(localAddress.sin_port)
  ); 

  retval = system(rule);
  if (retval)
    panic(ERR_EXTCMD, "NetfilterProxy: Cannot create Netfilter rule: '%s' (exit code %d)", rule, retval);
  rulesEstablished ++;

  /* Insert capture rule: Receiver -> Remote */

  snprintf(rule, sizeof(rule) - 1,   
    "/sbin/iptables -I OUTPUT -p tcp -s %s -d %s --sport %d --dport %d -j QUEUE\n",  
    renderIP(ntohl(localAddress.sin_addr.s_addr), buf1), renderIP(ntohl(remoteAddress.sin_addr.s_addr), buf2),
    receiverPort, htons(remoteAddress.sin_port)
  );

  retval = system(rule);
  if (retval)
    panic(ERR_EXTCMD, "NetfilterProxy: Cannot create Netfilter rule: '%s' (exit code %d)", rule, retval);
  rulesEstablished ++;

  /* Print some statistics to the trace file */
  
  if (packetTrace) {
    packetTrace->addComment("Local IP: %s", renderIP(ntohl(localAddress.sin_addr.s_addr), buf1));
    packetTrace->addComment("Remote IP: %s", renderIP(ntohl(remoteAddress.sin_addr.s_addr), buf1));
    packetTrace->addComment("Sender Port %d Receiver Port %d Proxy Port %d Remote Port %d", senderPort, receiverPort, ntohs(localAddress.sin_port), ntohs(remoteAddress.sin_port));
  }

  initialized = true;
}

void NetfilterProxy::run()
{
  /* Open connection to ip_queue */

  ipqHandle = ipq_create_handle(0, PF_INET);
  if (!ipqHandle) 
    panic(ERR_LIBIPQ, "NetfilterProxy: Cannot create libipq handle");

  int recvBufSize = ipqBufferSizeBytes;
  int status = setsockopt(((ipq_handle*)ipqHandle)->fd, SOL_SOCKET, SO_RCVBUF, &recvBufSize, sizeof(recvBufSize)); 
  if (status < 0) 
    panic(ERR_LIBIPQ, "NetfilterProxy: Cannot set libipq buffer size");
  
  status = ipq_set_mode(((ipq_handle*)ipqHandle), IPQ_COPY_PACKET, ipqBufferSizeBytes);
  if (status < 0)
    panic(ERR_LIBIPQ, "NetfilterProxy: Cannot set mode libipq; maybe 'modprobe ip_queue'?");

  /* Main loop */

  dprintf("Proxy running\n");

  Packet *packet = new Packet();
  unsigned char buffer[2000];
  while (true) {
    bool retry;
    struct timeval tv;
    do {
      retry = false;
      int status = ipq_read(((ipq_handle*)ipqHandle), buffer, sizeof(buffer), 5000000);
      gettimeofday(&tv, NULL);
      if (status < 0)
        panic(ERR_CANTREAD, "NetfilterProxy: Cannot read from Netfilter queue: %s", ipq_errstr());
      if (status == 0) {
        if (errno == EINTR)
          retry = true;
        else if (errno == 0)
          panic(ERR_TIMEOUT, "NetfilterProxy: Timeout expired; maybe remote host is unreachable?");
        else 
          panic(ERR_LIBIPQ, "NetfilterProxy: Read from Netfilter queue returned zero (%s)", strerror(errno));
      }
    } while (retry);

    switch (ipq_message_type(buffer)) {
      case IPQM_PACKET:
      {
        ipq_packet_msg_t *msg = ipq_get_packet(buffer);
        assert(msg->data_len > 0);
        
        packet->copyFrom((unsigned char*)&msg->payload, msg->data_len);
        
        int status = ipq_set_verdict(((ipq_handle*)ipqHandle), msg->packet_id, NF_DROP, 0, NULL);
        if (status < 0)
          panic(ERR_LIBIPQ, "NetfilterProxy: Cannot set verdict on packet #%lu", msg->packet_id);
        
        break;
      }
      case NLMSG_ERROR:
        panic(ERR_LIBIPQ, "NetfilterProxy: Received error message %d from Netfilter", ipq_get_msgerr(buffer));
      default:
        panic(ERR_INTERNAL, "NetfilterProxy: Unknown message type %d", ipq_message_type(buffer));
    }
    
    if ((packet->getSource() == ntohl(remoteAddress.sin_addr.s_addr)) && (packet->getDestination() == ntohl(localAddress.sin_addr.s_addr))) {
      if (packetTrace)
        packetTrace->addPacket(&tv, packet);
      policy->receivePacketFromRemote(packet);
    } else if (packet->isSource(&senderAddress) && packet->isDestination(&remoteAddress)) {
      if (packetTrace)
        packetTrace->addPacket(&tv, packet);
      policy->receivePacketFromSender(packet);
    } else if (packet->isSource(&receiverAddress) && packet->isDestination(&remoteAddress)) {
      if (packetTrace)
        packetTrace->addPacket(&tv, packet);
      policy->receivePacketFromReceiver(packet);
    } else {
      fprintf(stderr, "Catching weird packets (maybe from a former monarch run?). Ignoring it.\n");
      fprintf(stderr, "%X:%d->%X:%d\n", packet->getSource(), packet->getSourcePort(), packet->getDestination(), packet->getDestinationPort());
    }
  }
}
;
