#include <netinet/in.h>
#include <netpacket/packet.h>
#include <net/ethernet.h>
#include <net/if.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <linux/if_tun.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <stdio.h>
#include <fcntl.h>
#include <errno.h>

#include "monarch.h"
#include "proxies.h"
#include "traces.h"

#define dprintf(a...) do { } while (0)
//#define dprintf(a...) printf(a)

TunProxy::TunProxy() : Proxy()
{
  this->networkDeviceName[0] = 0;
  this->initialized = false;
  this->tunHandle = -1;
  this->packetTrace = NULL;
  this->senderHostRouteActive = false;
  this->receiverHostRouteActive = false;
  this->captureSocket = -1;
  this->sendSocket = -1;
  memset(&addresses, 0, sizeof(addresses));
}

TunProxy::~TunProxy()
{
  if (senderHostRouteActive) {
    char cmd[200], buf1[200];
    sprintf(cmd, "/sbin/route del %s dev %s", renderIP(ntohl(addresses.senderPeerAddress.sin_addr.s_addr), buf1), tunDeviceName);
    int retval = system(cmd);
    if (retval != 0) 
      fprintf(stderr, "Cannot execute '%s', exit code is %d", cmd, retval);
    senderHostRouteActive = false;
  }

  if (receiverHostRouteActive) {
    char cmd[200], buf1[200];
    sprintf(cmd, "/sbin/route del %s dev %s", renderIP(ntohl(addresses.receiverPeerAddress.sin_addr.s_addr), buf1), tunDeviceName);
    int retval = system(cmd);
    if (retval != 0) 
      fprintf(stderr, "Cannot execute '%s', exit code is %d", cmd, retval);
    receiverHostRouteActive = false;
  }
  
  if (captureSocket >= 0) {
    close(captureSocket);
    captureSocket = -1;
  }
  
  if (sendSocket >= 0) {
    close(sendSocket);
    sendSocket = -1;
  }
  
  if (tunHandle >= 0) {
    close(tunHandle);
    tunHandle = -1;
  }
}

void TunProxy::setupTUN(struct addressBlock *addressArg) 
{
  char buf1[256], buf2[256];
  assert(tunHandle < 0);

  struct ifreq ifr;
  struct sockaddr_in *saddr;
  int sd; // socket for configuring the tun device

  /* Regarding /dev/net/tun: character device, major 10 minor 200. For Debian, the vtun package will set it up automatically. */
  if ((tunHandle = open("/dev/net/tun", O_RDWR)) < 0) 
    panic(ERR_NOTFOUND, "TunProxy: Cannot open /dev/net/tun (code %d). Is the tun module loaded and the device file available?", tunHandle);

  memset(&ifr, 0, sizeof(ifr));
  ifr.ifr_flags = IFF_TUN | IFF_NO_PI; /* TUN device; do not provide packet information */

  if (ioctl(tunHandle, TUNSETIFF, (void*) &ifr) < 0)
    panic(ERR_INTERNAL, "TunProxy: Cannot set TUN device interface (%s)", strerror(errno));

  strcpy(tunDeviceName, ifr.ifr_name);
  dprintf("Allocated tun device %s\n", tunDeviceName);
  int tun_num = atoi(&tunDeviceName[3]); // get number of the tun device to use it for IP address choice

  if ((sd = socket(PF_INET, SOCK_STREAM, 0)) < 0)
    panic(ERR_INTERNAL, "TunProxy: Cannot allocate socket (%s)", strerror(errno));

  /* Get the current flags */
  if (ioctl(sd, SIOCGIFFLAGS, &ifr) == -1) 
    panic(ERR_INTERNAL, "TunProxy: Failed to get TUN flags (%s)", strerror(errno));

  /* if_up */
  ifr.ifr_flags |= IFF_UP;
  if (ioctl(sd, SIOCSIFFLAGS, &ifr) == -1) 
    panic(ERR_INTERNAL, "TunProxy: Failed to set TUN flags (%s)", strerror(errno));

  /* Set MTU of tun device (most of the time not a good idea) */
  //  ifr.ifr_mtu = 16*1024;
  //  if (ioctl(sd, SIOCSIFMTU , &ifr) == -1)
  //    panic(ERR_INTERNAL, "TunProxy: Cannot set MTU of TUN device (%s)", strerror(errno));

  dprintf("Allocated tun #%d\n", tun_num);

  memset(&ifr, 0, sizeof(ifr));
  strcpy(ifr.ifr_name, tunDeviceName);

  struct sockaddr_in tunLocalAddress;
  tunLocalAddress.sin_family = AF_INET;

  addresses = *addressArg;

  // The tun address can be specified by the --sender flag
  if(!addresses.senderAddress.sin_family) {
    tunLocalAddress.sin_addr.s_addr = htonl(0x0A000001 + (tun_num * 256));
    tunLocalAddress.sin_port = 0;
  }
  else{
    tunLocalAddress.sin_addr.s_addr = addresses.senderAddress.sin_addr.s_addr;
    tunLocalAddress.sin_port = 0;
  }

  memmove(&ifr.ifr_addr, &tunLocalAddress, sizeof(tunLocalAddress));
  if (ioctl(sd, SIOCSIFADDR, &ifr) == -1) 
    panic(ERR_INTERNAL, "TunProxy: Setting TUN device address failed (%s)", strerror(errno));

  /* Peers */
  
  uint16 ports[2];
  findUnusedPorts(tunLocalAddress.sin_addr.s_addr, ports, 2);
  
  if (!addresses.senderAddress.sin_family) {
    addresses.senderAddress = tunLocalAddress;
    addresses.senderAddress.sin_port = ports[0];
  }
  
  if (!addresses.receiverAddress.sin_family) {
    addresses.receiverAddress = tunLocalAddress;
    addresses.receiverAddress.sin_port = ports[1];
  }
  
  if (!addresses.senderPeerAddress.sin_family) {
    addresses.senderPeerAddress = tunLocalAddress;
    addresses.senderPeerAddress.sin_addr.s_addr = htonl(ntohl(addresses.senderPeerAddress.sin_addr.s_addr) + 1);
    addresses.senderPeerAddress.sin_port = htons(10001);
  }
  
  if (!addresses.receiverPeerAddress.sin_family) {
    addresses.receiverPeerAddress = tunLocalAddress;
    addresses.receiverPeerAddress.sin_addr.s_addr = htonl(ntohl(addresses.receiverPeerAddress.sin_addr.s_addr) + 1);
    addresses.receiverPeerAddress.sin_port = htons(10002);
  }
  
  /* Set up host route(s) */
  
  char cmd[200];
  sprintf(cmd, "/sbin/route add %s dev %s", renderIP(ntohl(addresses.senderPeerAddress.sin_addr.s_addr), buf1), tunDeviceName);
  int retval = system(cmd);
  if (retval != 0)
    panic(ERR_EXTCMD, "TunProxy: Command '%s' failed; exit code is %d", cmd, retval);
    
  senderHostRouteActive = true;
  
  if (addresses.receiverPeerAddress.sin_addr.s_addr != addresses.senderPeerAddress.sin_addr.s_addr) {
    char cmd[200];
    sprintf(cmd, "/sbin/route add %s dev %s", renderIP(ntohl(addresses.receiverPeerAddress.sin_addr.s_addr), buf1), tunDeviceName);
    int retval = system(cmd);
    if (retval != 0)
      panic(ERR_EXTCMD, "TunProxy: Command '%s' failed; exit code is %d", cmd, retval);
    
    receiverHostRouteActive = true;
  }  
}

void TunProxy::init(struct addressBlock *addressArg, const char *deviceName)
{
  char buf1[256];
  assert(!initialized && policy && addressArg->localAddress.sin_family && addressArg->remoteAddress.sin_family);
  
  if (geteuid() != 0)
    panic(ERR_PRIVILEGE, "TunProxy: Must be run as user 'root'");
  
  strncpy(networkDeviceName, deviceName, sizeof(networkDeviceName));
  policy->setProxy(this);
  setupTUN(addressArg);
  
  /* Open packet sockets */
  
  if ((captureSocket = socket(PF_PACKET, SOCK_DGRAM, htons(ETH_P_IP))) < 0)
    panic(ERR_INTERNAL, "TunProxy: Cannot open capture socket (%s)", strerror(errno));

  if ((sendSocket = socket(AF_INET, SOCK_RAW, IPPROTO_RAW)) < 0)
    panic(ERR_INTERNAL, "TunProxy: Cannot open send socket (%s)", strerror(errno));

  /* Get interface index for DEV */
  
  struct ifreq ifr;
  memset(&ifr, 0, sizeof(ifr));
  strncpy(ifr.ifr_name, networkDeviceName, sizeof(ifr.ifr_name));
  if (ioctl(captureSocket, SIOCGIFINDEX, &ifr) < 0)
    panic(ERR_INTERNAL, "TunProxy: Cannot get interface index for '%s' (%s)", networkDeviceName, strerror(errno));

  /* Bind capture socket to DEV */
  
  struct sockaddr_ll ll;
  memset(&ll, 0, sizeof(ll));
  ll.sll_family = AF_PACKET;
  ll.sll_ifindex = ifr.ifr_ifindex;
  ll.sll_protocol = htons(ETH_P_IP);
  if (bind(captureSocket, (struct sockaddr*)&ll, sizeof(ll)) < 0)
    panic(ERR_INTERNAL, "Cannot bind capture socket (%s)", strerror(errno));

  /* Set HDRINCL option on send socket */

  int flag = 1;
  if (setsockopt(sendSocket, IPPROTO_IP, IP_HDRINCL, (char*)&flag, sizeof(flag)) < 0)
    panic(ERR_INTERNAL, "Cannot set IP_HDRINCL option on send socket (%s)", strerror(errno));

  /* Print some statistics to the trace file */
  
  if (packetTrace) {
    packetTrace->addComment("Local IP: %s", renderIP(ntohl(addresses.localAddress.sin_addr.s_addr), buf1));
    packetTrace->addComment("Remote IP: %s", renderIP(ntohl(addresses.remoteAddress.sin_addr.s_addr), buf1));
    packetTrace->addComment("TUN IP: %s", renderIP(ntohl(addresses.senderAddress.sin_addr.s_addr), buf1));
    packetTrace->addComment("FAKE IP: %s", renderIP(ntohl(addresses.senderPeerAddress.sin_addr.s_addr), buf1));
    packetTrace->addComment("Sender Port %d Receiver Port %d Proxy Port %d Remote Port %d", ntohs(addresses.senderAddress.sin_port), ntohs(addresses.receiverAddress.sin_port), ntohs(addresses.localAddress.sin_port), ntohs(addresses.remoteAddress.sin_port));
  }

  initialized = true;
}

void TunProxy::getAddresses(struct sockaddr_in *senderLocal, struct sockaddr_in *senderRemote, struct sockaddr_in *receiverLocal, struct sockaddr_in *receiverRemote)
{
  if (senderLocal)
    *senderLocal = addresses.senderAddress;
  if (senderRemote)
    *senderRemote = addresses.senderPeerAddress;
  if (receiverLocal)
    *receiverLocal = addresses.receiverAddress;
  if (receiverRemote)
    *receiverRemote = addresses.receiverPeerAddress;
}

void TunProxy::sendRaw(Packet *p) 
{
  struct sockaddr_in sockAddr;
  int nbytes = 0;
  sockAddr.sin_family = AF_INET;
  sockAddr.sin_addr.s_addr = htonl(p->getDestination());
  sockAddr.sin_port = htons(p->getDestinationPort());
  struct timeval tv;
  gettimeofday(&tv, NULL);
  if (packetTrace)
    packetTrace->addPacket(&tv, p);
  nbytes = sendto(sendSocket, p->getBuffer(), p->getIpLength(), 0, (struct sockaddr *)&sockAddr, sizeof(sockAddr));
  if (nbytes < p->getIpLength())
    panic(ERR_CANTSEND, "TunProxy: Tried to write %d bytes raw packet, but sendto returned %d", p->getIpLength(), nbytes);
}

void TunProxy::forwardPacketToReceiver(Packet *packet)
{
  if (packet->isTcpPacket()) {
    uint8 *mssOption = packet->findTcpOption(Packet::TCP_OPTION_MSS);
    if (mssOption && (forcedMTU >= 0)) 
      *(uint16*)&mssOption[2] = htons(forcedMTU-(sizeof(struct IpHeader)+sizeof(struct TcpHeader)));
  }
  
  packet->setSourceAndDestination(&addresses.receiverPeerAddress, &addresses.receiverAddress);
  sendRaw(packet);
}

void TunProxy::forwardPacketToSender(Packet *packet)
{
  if (packet->isTcpPacket()) {
    uint8 *mssOption = packet->findTcpOption(Packet::TCP_OPTION_MSS);
    if (mssOption && (forcedMTU >= 0)) 
      *(uint16*)&mssOption[2] = htons(forcedMTU-(sizeof(struct IpHeader)+sizeof(struct TcpHeader)));
  }
  
  packet->setSourceAndDestination(&addresses.senderPeerAddress, &addresses.senderAddress);
  sendRaw(packet);
}

void TunProxy::sendPacketToRemote(Packet *packet)
{
  packet->setSourceAndDestination(&addresses.localAddress, &addresses.remoteAddress);
  sendRaw(packet);
}

void TunProxy::run()
{
  char buf1[256], buf2[256];
  verbose(1, "Probe exchange: %s <-> %s", renderAddress(&addresses.localAddress, buf1), renderAddress(&addresses.remoteAddress, buf2));
  verbose(1, "Sender: %s <-> %s", renderAddress(&addresses.senderAddress, buf1), renderAddress(&addresses.senderPeerAddress, buf2));
  verbose(1, "Receiver: %s <-> %s", renderAddress(&addresses.receiverAddress, buf1), renderAddress(&addresses.receiverPeerAddress, buf2));

  Packet *packet = new Packet();
  while (true) {
    fd_set fdr;
    FD_ZERO(&fdr);
    FD_SET(captureSocket, &fdr);
    FD_SET(tunHandle, &fdr);

    int maxHandle = max(captureSocket, tunHandle);
    
    int retval;
    do {
      retval = select(maxHandle+1, &fdr, NULL, NULL, NULL);
      if (retval < 0) {
        if (errno == EINTR)
          continue;
        panic(ERR_SELECT, "TunProxy: Select failed (%s)", strerror(errno));
      }
    } while (false);
    
    struct timeval tv;
    gettimeofday(&tv, NULL);
    
    if (FD_ISSET(captureSocket, &fdr)) {
      int retval = read(captureSocket, packet->getBuffer(), packet->getBufferLen());
      if (retval <= 0)
        panic(ERR_CANTREAD, "TunProxy: Cannot read from capture socket: %d", retval);

      packet->setLength(retval);
      
      // Note: This is a fairly simple filter and will not work if you have more than one
      // Monarch flow to the same destination host. TODO: Implement a better filter.
      
      if ((packet->getSource() == ntohl(addresses.remoteAddress.sin_addr.s_addr)) && (packet->getDestination() == ntohl(addresses.localAddress.sin_addr.s_addr))) {

        if (packetTrace) 
          packetTrace->addPacket(&tv, packet);
        policy->receivePacketFromRemote(packet);
      }
    }
    
    if (FD_ISSET(tunHandle, &fdr)) {
      int retval = read(tunHandle, packet->getBuffer(), packet->getBufferLen());
      if (retval <= 0)
        panic(ERR_CANTREAD, "TunProxy: Cannot read from TUN device: %d", retval);

      packet->setLength(retval);

      if (packet->isSource(&addresses.senderAddress) && packet->isDestination(&addresses.senderPeerAddress)) {
        if (packetTrace)
          packetTrace->addPacket(&tv, packet);
        policy->receivePacketFromSender(packet);
      } else if (packet->isSource(&addresses.receiverAddress) && packet->isDestination(&addresses.receiverPeerAddress)) {
        if (packetTrace)
          packetTrace->addPacket(&tv, packet);
        policy->receivePacketFromReceiver(packet);
      } else {
        fprintf(stderr, "Caught an unexpeced packet on TUN (%X:%d->%X:%d); ignoring\n", 
          packet->getSource(), packet->getSourcePort(), packet->getDestination(), packet->getDestinationPort()
        );
      }
    }
  }
}
