#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <arpa/inet.h>
#include <sys/socket.h>
#include <sys/time.h>

#include "monarch.h"
#include "traces.h"

#define dprintf(a...) do { } while (0)
//#define dprintf(a...) printf(a)

Receiver::Receiver(struct sockaddr_in *localAddress, struct sockaddr_in *remoteAddress)
{
  this->localAddress = *localAddress;
  this->remoteAddress = *remoteAddress;
  this->receiveBufferSize = 0;
  this->readGranularity = 1*MB;
  this->initialized = false;
}

void Receiver::init() 
{
  char buf1[256], buf2[256];
  assert(!initialized);

  dprintf("Receiver: Configured for %s->%s\n", renderAddress(&localAddress, buf1), renderAddress(&remoteAddress, buf2));

  /* Create a socket for the receiver */
  
  myServerSocket = socket(AF_INET, SOCK_STREAM, 0);
  if (myServerSocket < 0) 
    panic(1, "Receiver: Cannot create socket (%s)", strerror(errno));
  
  /* Configure the socket */
  
  if (receiveBufferSize) {
    if (setsockopt(myServerSocket, SOL_SOCKET, SO_RCVBUF, &receiveBufferSize, sizeof(receiveBufferSize)))
      panic(2, "Receiver: Cannot set receive buffer size to %d bytes", receiveBufferSize);

    /* Sanity checks */
    int actualSize;
    socklen_t optlen2 = sizeof(actualSize);
    if (getsockopt(myServerSocket, SOL_SOCKET, SO_RCVBUF, &actualSize, &optlen2) >= 0)
      verbose(1, "Receive buffer changed to %d bytes", actualSize);
  }
  else{
    /* Sanity checks */
    int actualSize;
    socklen_t optlen2 = sizeof(actualSize);
    if (getsockopt(myServerSocket, SOL_SOCKET, SO_RCVBUF, &actualSize, &optlen2) >= 0)
      dprintf("Receive buffer has size %d bytes\n", actualSize);
  }

  int yes = 1;
  if (setsockopt(myServerSocket, SOL_SOCKET, SO_REUSEADDR, &yes, sizeof(yes)))
    panic(2, "Receiver: Cannot set SO_REUSEADDR option");

  /* Allocate receive buffer */

  buffer = (char*) malloc(readGranularity);
  if (!buffer)
    panic(3, "Receiver: Cannot allocate a buffer of %d bytes", readGranularity);
 

  /* Bind to the socket, and set it to listen */

  if (bind(myServerSocket, (struct sockaddr *) &localAddress, sizeof(localAddress)) < 0)
    panic(4, "Receiver: Cannot bind socket to %s (%s)", renderAddress(&localAddress, buf1), strerror(errno));

  const int queueSize = 1;
  if (listen(myServerSocket, queueSize) < 0)
    panic(5, "Receiver: Cannot set socket to listen mode (%s)", strerror(errno));

  dprintf("Receiver ready to accept\n");
  initialized = true;
}

void Receiver::run()
{
  assert(initialized);

  /* Accept incoming connection */
  
  struct sockaddr peerAddress;
  socklen_t peerAddressLength = sizeof(peerAddress); 
  int mySocket = accept(myServerSocket, &peerAddress, &peerAddressLength);
  if (mySocket < 0)
    panic(6, "Receiver: Error while accepting connection (%s)", strerror(errno));

  dprintf("Receiver accepted\n");

  /* Read whatever the sender is sending */
  
  bool connectionClosed = false;
  bool senderTerminated = false;
  struct timeval nextSnapshot;
  struct tcp_info tcpState;
  gettimeofday(&nextSnapshot, NULL);
  
  while (true) {
    struct timeval tv;
    gettimeofday(&tv, NULL);
    while (isNotLater(nextSnapshot, tv)) {
      socklen_t len = sizeof(tcpState);
      if (getsockopt(mySocket, SOL_TCP, TCP_INFO, &tcpState, &len) < 0)
        panic(1, "Receiver: Cannot get TCP_INFO for socket (%s)", strerror(errno));
      if (kernelTrace)
        kernelTrace->addSnapshot(&tv, &tcpState);
      if (tcpState.tcpi_state == TCP_CLOSE) 
        connectionClosed = true;
        
      nextSnapshot.tv_usec += 10000;
      while (nextSnapshot.tv_usec > 1000000) {
        nextSnapshot.tv_usec -= 1000000;
        nextSnapshot.tv_sec ++;
      }
    }

    if (connectionClosed)
      break;

    tv.tv_sec = nextSnapshot.tv_sec - tv.tv_sec;
    if (nextSnapshot.tv_usec > tv.tv_usec) {
      tv.tv_usec = nextSnapshot.tv_usec - tv.tv_usec;
    } else {
      tv.tv_usec = (nextSnapshot.tv_usec + 1000000) - tv.tv_usec;
      tv.tv_sec -= 1;
    }

    fd_set fdr;
    FD_ZERO(&fdr);
    if (!senderTerminated)
      FD_SET(mySocket, &fdr);
      
    int retval = select(mySocket+1, NULL, &fdr, NULL, &tv);

    if (FD_ISSET(mySocket, &fdr)) {
      int bytesRead = read(mySocket, buffer, readGranularity);
      if (bytesRead < 0)
        panic(7, "Receiver: Cannot read from socket (%s)", strerror(errno));
      if (bytesRead == 0) {
        shutdown(mySocket, SHUT_RDWR);
        senderTerminated = true;
      }
      dprintf("Read %d bytes\n", bytesRead);
    }
  }

  /* More sanity checks */

  int actualSize;
  socklen_t optlen2 = sizeof(actualSize);
  if (getsockopt(mySocket, SOL_SOCKET, SO_RCVBUF, &actualSize, &optlen2) >= 0)
    dprintf("End: %d bytes receive buffer\n", actualSize);  
  
  /* Terminate the connection */

  if (close(mySocket) < 0)
    panic(9, "Receiver: Cannot terminate connection (%s)", strerror(errno));

  dprintf("Receiver shutdown.\n");
  free(buffer);
}
