#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <sys/socket.h>
#include <sys/time.h>

#include "monarch.h"
#include "traces.h"

#define dprintf(a...) do { } while (0)
//#define dprintf(a...) printf(a)

Sender::Sender(struct sockaddr_in *localAddress, struct sockaddr_in *remoteAddress)
{
  this->localAddress = *localAddress;
  this->remoteAddress = *remoteAddress;
  this->writeGranularity = 1*MB;
  this->messageFileName = NULL;
  this->sendBufferSize = 0;
  this->transferSize = 500*KB;
  this->transferEnd.tv_sec = 0;
  this->transferEnd.tv_usec = 0;
}

void Sender::run() 
{
  char buf1[256], buf2[256];
  dprintf("Sender: Configured for %s->%s\n", renderAddress(&localAddress, buf1), renderAddress(&remoteAddress, buf2));

  /* Create a socket, and bind it to the specified local address */
  
  int mySocket = socket(AF_INET, SOCK_STREAM, 0);
  if (mySocket < 0)
    panic(ERR_INTERNAL, "Sender: Cannot create socket (%s)", strerror(errno));

  if (bind(mySocket, (struct sockaddr *) &localAddress, sizeof(localAddress)) < 0)
    panic(ERR_ADDRINUSE, "Sender: Cannot bind to %s (%s)", renderAddress(&localAddress, buf1), strerror(errno));

  /* Set up the user-level send buffer, and fill it with a message
     describing the traffic */
  
  char *buffer = (char*)malloc(writeGranularity);
  if (!buffer)
    panic(ERR_INTERNAL, "Sender: Cannot allocate %d byte write buffer", writeGranularity);
    
  int messageLength;
  if (messageFileName) {
    FILE *messageFile = fopen(messageFileName, "r");
    if (!messageFile)
      panic(ERR_NOTFOUND, "Sender: Cannot open message file '%s'", messageFileName);
    
    messageLength = fread(buffer, 1, writeGranularity, messageFile);
    if (messageLength < 1)
      panic(5, "Sender: Cannot read from message file '%s'", messageFileName);
      
    fclose(messageFile);
  } else {
    char *user = getenv("USER");
    char *hostname = getenv("HOSTNAME");
    messageLength = snprintf(buffer, writeGranularity, "This measurement traffic is being generated by %s@%s using %s v%s. Please contact the sender for more information.", 
	    user ? user : "unknownUser", hostname ? hostname : "unknownHost", PACKAGE, VERSION
    );
  }
  
  for (int i=messageLength; i<writeGranularity; i++)
    buffer[i] = buffer[i % messageLength];

  /* Configure the socket. This must be done BEFORE connecting! */

  if (sendBufferSize > 0) {
    if (setsockopt(mySocket, SOL_SOCKET, SO_SNDBUF, &sendBufferSize, sizeof(sendBufferSize)))
      panic(6, "Sender: Cannot set send buffer size to %d bytes", sendBufferSize);

    /* Sanity checks */
    int actualSize;
    socklen_t optlen2 = sizeof(actualSize);
    if (getsockopt(mySocket, SOL_SOCKET, SO_SNDBUF, &actualSize, &optlen2) >= 0)
      verbose(1, "Send buffer changed to %d bytes", actualSize);
  }
  else{
    /* Sanity checks */
    int actualSize;
    socklen_t optlen2 = sizeof(actualSize);
    if (getsockopt(mySocket, SOL_SOCKET, SO_SNDBUF, &actualSize, &optlen2) >= 0)
      dprintf("Send buffer has size %d bytes\n", actualSize);
  }
  
  /* Try to connect to the receiver */

  dprintf("Sender attempting to connect\n");
  if (connect(mySocket, (struct sockaddr *)&remoteAddress, sizeof(remoteAddress)) < 0)
    panic(7, "Sender: Cannot connect to receiver (%s)", strerror(errno));

  dprintf("Sender connected\n");

  /* Set O_NONBLOCK */
  
  int flags = fcntl(mySocket, F_GETFL);
  fcntl(mySocket, F_SETFL, flags | O_NONBLOCK);

  /* Sanity checks */

  int actualSize;
  socklen_t optlen2 = sizeof(actualSize);
  if (getsockopt(mySocket, SOL_SOCKET, SO_SNDBUF, &actualSize, &optlen2) >= 0)
    dprintf("Start: %d bytes send buffer\n", actualSize);  

  /* Do the actual transfer */

  int bytesRemaining = transferSize;
  struct timeval nextSnapshot;
  int writePointer = 0;
  struct tcp_info tcpState;
  bool connectionClosed = false;
  gettimeofday(&nextSnapshot, NULL);
  
  while ((bytesRemaining > 0) || !connectionClosed) {
    int bytesToWrite = min(writeGranularity - writePointer, bytesRemaining);
    int bytesWritten;
    int retval;

    struct timeval tv;
    struct timeval now;
    gettimeofday(&tv, NULL);
    now.tv_sec = tv.tv_sec;
    now.tv_usec = tv.tv_usec;

    while (isNotLater(nextSnapshot, tv)) {
      socklen_t len = sizeof(tcpState);
      if (getsockopt(mySocket, SOL_TCP, TCP_INFO, &tcpState, &len) < 0)
        panic(1, "Sender: Cannot get TCP_INFO for socket (%s)", strerror(errno));
      if (tcpState.tcpi_state == TCP_CLOSE)
        connectionClosed = true;
      if (kernelTrace)
        kernelTrace->addSnapshot(&tv, &tcpState);

      nextSnapshot.tv_usec += 10000;
      while (nextSnapshot.tv_usec > 1000000) {
        nextSnapshot.tv_usec -= 1000000;
        nextSnapshot.tv_sec ++;
      }
    }

    tv.tv_sec = nextSnapshot.tv_sec - tv.tv_sec;
    if (nextSnapshot.tv_usec > tv.tv_usec) {
      tv.tv_usec = nextSnapshot.tv_usec - tv.tv_usec;
    } else {
      tv.tv_usec = (nextSnapshot.tv_usec + 1000000) - tv.tv_usec;
      tv.tv_sec -= 1;
    }

    fd_set fdw;
    FD_ZERO(&fdw);
    if (bytesToWrite > 0)
      FD_SET(mySocket, &fdw);
      
    retval = select(mySocket+1, NULL, &fdw, NULL, &tv);
 
    if (FD_ISSET(mySocket, &fdw)) {
      bytesWritten = write(mySocket, &buffer[writePointer], bytesToWrite);
      if (bytesWritten <= 0)
        panic(8, "Sender: Cannot write %d bytes to socket (%d bytes remaining)", bytesToWrite, bytesRemaining);

      writePointer += bytesWritten;
      bytesRemaining -= bytesWritten;
    
      if (writePointer >= writeGranularity)
        writePointer = 0;

      /* Timed transfer */
      if(this->transferEnd.tv_sec > 0) {
	if(isNotLater(now, this->transferEnd))
	  bytesRemaining = transferSize; // Reset to never run out of data
	else {
	  verbose(1, "Sender: Ending transfer because the time is up (%ds %dms more)", now.tv_sec - this->transferEnd.tv_sec, (now.tv_usec-this->transferEnd.tv_usec)/1000);
	  bytesRemaining = 0;
	}
      }
        
      if (bytesRemaining < 1) {
        if (shutdown(mySocket, SHUT_WR) < 0)
          panic(8, "Sender: Cannot terminate connection");
      }
    }
  }
  /* More sanity checks */

  optlen2 = sizeof(actualSize);
  if (getsockopt(mySocket, SOL_SOCKET, SO_SNDBUF, &actualSize, &optlen2) >= 0)
    dprintf("End: %d bytes send buffer\n", actualSize);  

  /* Close the connection */

  if (close(mySocket) < 0) 
    panic(8, "Sender: Cannot terminate connection");

  dprintf("Sender finished\n");
  free(buffer);
}

